<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>{{ title }}</title>
    <style>
        div, li {
            padding: 2px;
            font-family: monospace, sans-serif;
            font-weight: 400;
            font-style: normal;
        }

        h1, h2, h3 {
            padding: 2px;
            font-family: monospace, sans-serif;
            font-style: normal;
        }

        summary {
            padding: 2px;
            font-family: monospace, sans-serif;
            font-weight: 400;
            font-style: normal;
        }

        a {
            color: #fbfbfb;
        }

        .failed, .failed summary, .failed a {
            color: #ff8282;
        }

        .failed a {
            text-decoration: none;
        }

        .failed details {
            color: #dddddd;
        }

        .block {
            padding-left: 40px
        }

        ul {
            padding-left: 40px
        }

        body {
            background-color: #6c6c6c;
            color: #dddddd;
        }
    </style>
</head>
<body>
{% if page_type == 'simple_page' %}
{{ body }}
{% elif page_type == 'top_scenarios' %}
    <div>Список автотестов по времени выполнения <sup>(pt - процессорное время)</sup>:</div>
    <ul>
    {% for tc in testcases %}
        <li class="{{ tc.status }}">{{ tc.time }} sec. <sup>(pt {{ tc.time_process }} sec.)</sup>
        {% if tc.url %}
            <a href="{{ tc.url }}" target="blank">{{ tc.name }}</a>
        {% else %}
            {{ tc.name }}
        {% endif %}
            <sup>pid {{ tc.pid }}</sup>
        </li>
    {% endfor %}
    </ul>
{% elif page_type == 'autotest_result' %}
    <h2>Версия пакета</h2>
        {% if version_info %}
            <ul>
            {% for info in version_info %}
                <li>{{ info.verbose_name }} - {{ info.version }}</li>
            {% endfor %}
            </ul>
        {% else %}
            <div class="block">Нет данных</div>
        {% endif %}
    <h2>Результаты выполнения автотестов:</h2>
    <ul>
        <li>Всего выполнено автотестов: {{ total_count }}</li>
        {% if failed_data %}
        <li>Всего упавших автотестов: <b class="failed">{{ failed_data|length }}</b>
            <ul>
                {% for tc in failed_data %}
                    <li class="failed"><a href="#{{ tc.attrib.main_tag }}">{{ tc.attrib.name }}</a></li>
                {% endfor %}
            </ul>
        </li>
        {% endif %}
    </ul>
    <hr/>

    <div class="block">
    <h3>Выполненные автотесты по блокам:</h3>
    {% for suite_data in suites_data %}
        <details>
            <summary>{{ suite_data.suite.name }}</summary>
            <ul>
                {% for tc in suite_data.test_cases %}
                    <li class="{{ tc.attrib.status }}" id="{{ tc.attrib.main_tag }}">
                    {% if tc.attrib.status == 'passed' %}
                        {{ tc.attrib.name }}
                    {% elif tc.attrib.status == 'failed' %}
                        <details>
                            <summary>{{ tc.attrib.name }}</summary>
<pre>{{ tc.attrib.failure_text }}</pre>
                        </details>
                    {% endif %}
                    </li>
                {% endfor %}
            </ul>
        </details>
    {% endfor %}
    </div>
{% endif %}
</body>
</html>